"""
This file is for all xpath-related functions that are relevant to all means of web crawling.
"""


def lowercase(text: str):
    """
    Returns the xpath needed to translate a string to lowercase
    """
    return f"translate('{text}', 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', 'abcdefghijklmnopqrstuvwxyz')"


def contains_keywords(
    keywords: list[str],
    container: str = "text()",
    case_sensitive: bool = False,
):
    """
    Returns the xpath needed to search a container for keywords. Final xpath will look like:
    contains(text(), 'keyword1') and contains(text(), 'keyword2') and etc...
    Also, ensures keywords are searched for case-insensitively.
    """
    # This is to prevent user from having to add those quotes manually.
    # INCORRECT: contains(..., keyword), CORRECT: contains(..., 'keyword')
    ctn = container if container == "text()" else f"'{container}'"

    xpaths = []
    for kw in keywords:
        container_text = lowercase(ctn) if case_sensitive else ctn
        keyword_text = lowercase(kw) if case_sensitive else kw
        # Add quotes around keyword for same reason as above.
        keyword_text = f"'{keyword_text}'"
        xpath = f"contains({container_text}, {keyword_text})"
        xpaths.append(xpath)

    return " and ".join(xpaths)



# Standard Libraries
import json
import os
from urllib.parse import urlencode

# Third-Party Libraries
# * OPTIONAL Imports
try:
    from fake_useragent import FakeUserAgent
except ModuleNotFoundError:
    FakeUserAgent = None

# Local libraries
from .exceptions import FakeUserAgentNotInstalled


def add_params_to_url(url: str, params: dict) -> str:
    return f"{url}?{urlencode(params)}"


def get_cookie_header(cookies: list[dict]):
    return "; ".join([f"{cookie['name']}={cookie['value']}" for cookie in cookies])


def get_random_user_agent() -> str:
    if FakeUserAgent is None:
        raise FakeUserAgentNotInstalled
    ua = FakeUserAgent()
    return ua.random


def get_schema_from_json(data: dict, output: str = "schema.json", sample_size: int = 3):
    """
    Outputs schema of the data given a json object.
    @output: file to output schema to
    @sample_size: number of data samples
    """
    directory = os.path.dirname(output)

    if "/" in output and not os.path.exists(directory):
        # Create the directory if it DOESN'T exist
        os.makedirs(directory)

    with open(output, "w") as f:
        sample = data[:sample_size]
        json.dump(sample, f, indent=4)
        f.close()



from lytils import ctext


class FakeUserAgentNotInstalled(Exception):
    # Raise this when undetected_chromedriver is not installed
    def __init__(
        self,
        message="<y>Module 'fake-useragent' not installed.",
    ):
        self.message = ctext(message)
        super().__init__(self.message)



# Standard Libraries
import asyncio
import time

# Third-Party Libraries
from playwright.async_api import Page


async def delay_request(time_of_last_request: float, delay: float = 1):
    """
    To prevent sending too many requests to server at once.
    """
    elapsed_time = time.time() - time_of_last_request
    if elapsed_time < delay:
        await asyncio.sleep(delay - elapsed_time)
    return time.time()


async def get_ip(page: Page):
    await page.goto("https://www.whatismyip.com/")
    ip_element = await page.wait_for_selector(".ip-address")
    ip_address = await ip_element.text_content()
    return ip_address.strip()



# Standard Libraries
import random
import time

# Third-party Libraries
from lytils import cprint, cinput, print_trace, pause
from selenium.common.exceptions import TimeoutException, WebDriverException
from selenium.webdriver import Chrome, ChromeOptions
from selenium.webdriver import DesiredCapabilities
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from webdriver_manager.chrome import ChromeDriverManager

# * OPTIONAL IMPORTS
try:
    import undetected_chromedriver as uc
except ModuleNotFoundError:
    uc = None

try:
    from seleniumwire import webdriver as sw
except ModuleNotFoundError:
    sw = None

# Local Libraries
# Make it so installing browsermobproxy is optional
from .browsermob_proxy_wrapper import BrowsermobProxyWrapper
from .exceptions import SeleniumWireAndUndetectedChromedriverIncompatible
from .exceptions import SeleniumWireNotInstalled
from .exceptions import UndetectedChromedriverNotInstalled


class QSelenium:
    # Latest release can be found at:
    # https://chromedriver.storage.googleapis.com/
    _DEFAULT_CHROMEDRIVER_VERSION = "114.0.5735.90"

    def __init__(
        self,
        chrome_options: dict,
        chrome_capabilities: dict = {},
        chromedriver_path: str = "",
        chromedriver_version: str = "",
        clear_cookies: bool = False,
        bmp_options: dict = {},  # browsermob proxy
        use_undetected_chromedriver: bool = False,
        use_selenium_wire: bool = False,
        initialize_driver: bool = True,
        timeout_default: int = 10,
        debug: bool = False,
    ):
        # * Raise errors IMMEDIATELY if desired imports are not available
        self._use_uc = use_undetected_chromedriver
        if self._use_uc and uc == None:
            raise UndetectedChromedriverNotInstalled
        self._use_sw = use_selenium_wire
        if self._use_sw and sw == None:
            if self._use_uc:
                raise SeleniumWireAndUndetectedChromedriverIncompatible
            elif sw == None:
                raise SeleniumWireNotInstalled

        self._debug = debug
        self._timeout_default = timeout_default

        # * Set DEFAULT headers here
        self._headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
            "sec-ch-ua": '"Not_A Brand";v="8", "Chromium";v="120", "Brave";v="120"',
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
            "accept-language": "en-US;en;q=0.9",
            "accept-encoding": "gzip, deflate, br",
        }

        # Use browsermob proxy for scraping network traffic
        self._bmp = None
        if bmp_options:
            self.initialize_bmp(bmp_options)

        # Set chrome options
        self._chrome_options = None
        self._set_chrome_options(chrome_options)

        # Set chrome capabilities
        self._chrome_capabilities = None
        self._set_chrome_capabilities(chrome_capabilities)

        # Set chromedriver path
        self._chromedriver_path = None
        self._set_chromedriver_path(chromedriver_path, chromedriver_version)

        # Automatically initialize the driver by default on object creation.
        self._dr = None
        if initialize_driver:
            self.initialize_driver()

            # Try this to avoid tracking cookies potentially flagging your bot.
            if clear_cookies:
                self._clear_cookies()

            # Set real request headers for more sleuth
            self.set_headers()

        self._actions = None  # Use action chains

        # End of object initialization.

    def close(self):
        """
        Teardown Function
        """
        # Close driver if it was initialized
        if self._dr:
            self._dr.close()
            self._dr.quit()

        # Close browsermob-proxy if it exists
        if self._bmp:
            self._bmp.close()

    # * PRIVATE Functions START
    # region
    def _get_seleniumwire_options(self, proxy_url: str):
        # Return the seleniumwire_options to be used with seleniumwire.webdriver.Chrome
        return {
            "proxy": {
                "http": f"{proxy_url}",
                "https": f"{proxy_url}",
                "verify_ssl": True,
            },
        }

    def _set_chromedriver_path(self, chromedriver_path: str, chromedriver_version: str):
        # Get chromedriver version
        version = ""
        if chromedriver_version != "":
            version = chromedriver_version
        else:
            version = QSelenium._DEFAULT_CHROMEDRIVER_VERSION

        # Latest release can be found at:
        # https://chromedriver.storage.googleapis.com/
        if self._debug:
            cprint(f"<y>DEBUG: <w>Using chromedriver version {version}")

        if chromedriver_path != "":
            self._chromedriver_path = chromedriver_path
        else:
            # Automatically installs chromedriver with specified version
            try:
                self._chromedriver_path = ChromeDriverManager(version=version).install()
            except:
                # Alternative parameter name: driver_version
                self._chromedriver_path = ChromeDriverManager(
                    driver_version=version
                ).install()

    def _set_chrome_options(self, chrome_options: dict):
        o = ChromeOptions()

        # Adding argument to disable the AutomationControlled flag
        o.add_argument("--disable-blink-features=AutomationControlled")

        # Exclude the collection of enable-automation switches
        o.add_experimental_option("excludeSwitches", ["enable-automation"])

        # Turn-off userAutomationExtension
        o.add_experimental_option("useAutomationExtension", False)

        if self._use_uc:
            o = uc.ChromeOptions()
        else:
            # * Options that don't work in undetected chromedriver
            # Keep browser open after completion so we can debug issues
            if chrome_options.get("detach", False):
                o.add_experimental_option("detach", True)

            # Should remove unnecessary errors and warnings
            o.add_experimental_option("excludeSwitches", ["enable-logging"])

        # Maximize browser
        o.add_argument("--start-maximized")

        # Automatically block notifications
        o.add_argument("--disable-notifications")

        # Don"t want extensions to mess up bot
        o.add_argument("--disable-extensions")
        o.add_argument("--disable-default-apps")
        # o.add_argument('--ignore-certificate-errors')
        # o.add_argument('--ignore-ssl-errors')
        # Adding to fix graphical glitch with chrome
        o.add_argument("--disable-gpu")

        # o.add_argument('--disable-site-isolation-trials')
        # o.add_argument('--no-sandbox')
        # o.add_argument('--disable-dev-shm-usage')
        # o.add_argument("--no-default-browser-check")
        # o.add_argument("--no-first-run")
        # o.binary_location = 'G:\My Drive\Programs\Google Chrome\GoogleChromePortable\App\Chrome-bin\chrome.exe'

        if chrome_options.get("user_data_path"):
            o.add_argument(f'--user-data-dir={chrome_options["user_data_path"]}')
        if chrome_options.get("profile_path"):
            o.add_argument(f'--profile-directory={chrome_options["profile_path"]}')
        if chrome_options.get("headless"):
            o.add_argument(f"--headless")

        # Utilizing browsermob proxy
        if self._bmp is not None:
            o.add_argument(f"--proxy-server={self._bmp.proxy_url()}")

        if chrome_options.get("chrome_exe_path"):
            o.binary_location = chrome_options["chrome_exe_path"]

        self._chrome_options = o
        return self._chrome_options

    def _set_chrome_capabilities(self, chrome_capabilities: dict):
        c = DesiredCapabilities.CHROME.copy()

        # capabilities["acceptSslCerts"] = True

        # Use when receiving the chrome error: ERR_CERT_AUTHORITY_INVALID
        # capabilities["acceptInsecureCerts"] = True
        # capabilities['pageLoadStrategy'] = "normal", "eager", "none"

        c.update(chrome_capabilities)

        # ? Adjust cache for browser to store more data as cache and for longer
        # capabilities["browser_cache"] = {"enabled": True, "capacity": 10000000, "timeToLiveInSeconds": 300}
        self._chrome_capabilities = c
        return self._chrome_capabilities

    def _remove_navigator_webdriver(self):
        """
        Sets navigator.webdriver to undefined to avoid detection.
        """
        self._dr.execute_script(
            "Object.defineProperty(navigator, 'webdriver', {get: () => undefined})"
        )

    def _set_driver(self):
        chrome_params = {"options": self._chrome_options}
        if self._use_sw and self._bmp is not None:
            chrome_params.update(
                {
                    "seleniumwire_options": self._get_seleniumwire_options(
                        f"http://{self._bmp.proxy_url()}"
                    ),
                }
            )

        if self._use_uc:
            chrome_params.update(
                {
                    "driver_executable_path": self._chromedriver_path,
                    "browser_executable_path": self._chrome_options.binary_location,
                }
            )
        else:
            chrome_params.update({"service": Service(self._chromedriver_path)})

        if self._use_uc:
            self._dr = uc.Chrome(**chrome_params)
        elif self._use_sw:
            self._dr = sw.Chrome(**chrome_params)
        else:
            self._dr = Chrome(**chrome_params)

        self._remove_navigator_webdriver()

    def _clear_cookies(self):
        self._dr.delete_all_cookies()

    # endregion
    # * PRIVATE Functions END

    # Some websites may require specific cookies in order to categorize you as a real user.
    # To pull existing cookies, follow instructions on https://github.com/GGLionCross/CookieConverter
    # This must be called on the correct domain, or you'll get an InvalidCookieDomainException
    def add_cookies(self, cookies: list[dict]):
        """
        Some websites may require specific cookies in order to categorize you as a real user.
        To pull existing cookies, follow instructions on https://github.com/GGLionCross/CookieConverter
        This must be called on the correct domain, or you'll get an InvalidCookieDomainException
        """
        for cookie in cookies:
            self._dr.add_cookie(cookie)

    def initialize_driver(self):
        try:
            self._set_driver()
        except WebDriverException:
            warning = "Google profile may be in use. If so, please close and press <enter> to continue."
            print_trace()
            cinput(f"<y>{warning}")
            self._set_driver()
        except Exception:
            print_trace()

        return self._dr

    def get_driver(self):
        return self._dr

    def set_headers(self, headers: dict):
        self._headers.update(headers)

        def interceptor(request):
            # Delete previous headers to prevent duplicates
            # Then set real request headers
            for key, value in self._headers.items():
                del self._headers[key]
                self._headers[key] = value

        # Set the Selenium Wire interceptor
        self._dr.request_interceptor = interceptor

    def get_headers(self):
        return self._headers

    # * BROWSERMOB PROXY Functions START
    # region
    """
    Useful for scraping network traffic instead of web elements
    Using network traffic to scrape is often better b/c network traffic is
    less likely to change versus front-end design
    """

    def initialize_bmp(self, bmp_options: dict = {}):
        self._bmp = BrowsermobProxyWrapper(**bmp_options)
        self._bmp.start_server()
        self._bmp.start_client()
        return self._bmp

    def get_bmp(self):
        return self._bmp

    # endregion
    # * BROWSERMOB PROXY Functions END

    # * DEBUG Functions START
    # region
    def pause(self):
        pause()

    def screenshot(self, filename: str = "screenshot.png"):
        self._dr.get_screenshot_as_file(filename)

    # endregion
    # * DEBUG Functions END

    def use_actions(self):
        self._actions = ActionChains(self._dr)
        return self._actions

    # * DELAY Functions START
    # region
    def random_delay(self, min_delay: int, max_delay: int):
        time.sleep(random.uniform(min_delay, max_delay))

    # Delay functions will be useful so we're not traversing sites at unrealistic speeds
    def visit_with_delay(self, url: str, min_delay: int = 1, max_delay: int = 3):
        # Add delay when visiting url
        self.random_delay(min_delay, max_delay)
        self._dr.get(url)

    def exponential_backoff(self, url: str, max_retries: int = 3, wait_time: int = 1):
        """
        Useful when dealing with multiple requests to a site's server
        @param max_retries: Maximum number of retries to load a page
        @param wait_time: Initial wait time (in seconds) which gets doubled after each failure
        """
        time.sleep(wait_time)
        for i in range(max_retries):
            try:
                self._dr.get(url)
                break  # Break out of for loop if url is successful
            except TimeoutException as e:
                print(f"Failed getting url ({i + 1}): {e}")

                # If error occurs, wait_time is exponentially increased
                wait_time *= 2
                time.sleep(wait_time)

    # endregion
    # * DELAY Functions END

    # * WAIT Functions START
    # region Wait Functions
    def wait_until(self, condition, timeout=None):
        if timeout is None:
            timeout = self._timeout_default
        return WebDriverWait(self._dr, timeout).until(condition)

    def wait_until_clickable(self, descriptor, timeout=None):
        if timeout is None:
            timeout = self._timeout_default
        return WebDriverWait(self._dr, timeout).until(
            EC.element_to_be_clickable(descriptor)
        )

    def wait_for_element(self, descriptor, timeout=None):
        if timeout is None:
            timeout = self._timeout_default
        return WebDriverWait(self._dr, timeout).until(
            EC.presence_of_element_located(descriptor)
        )

    def wait_for_all_elements(self, descriptor, timeout=None):
        if timeout is None:
            timeout = self._timeout_default
        return WebDriverWait(self._dr, timeout).until(
            EC.presence_of_all_elements_located(descriptor)
        )

    def wait_for_text_not_empty(self, descriptor, timeout=None):
        if timeout is None:
            timeout = self._timeout_default
        WebDriverWait(self._dr, timeout).until(
            lambda driver: driver.find_element(*descriptor).text.strip() != ""
        )
        return WebDriverWait(self._dr, timeout).until(
            EC.presence_of_element_located(descriptor)
        )

    # endregion
    # * WAIT Functions END



from lytils import ctext


# BMP Exceptions
class BrowsermobProxyNotInstalled(Exception):
    # Raise this when browsermob proxy path is missing
    def __init__(
        self,
        message="Module 'browsermob-proxy' not installed.",
    ):
        self.message = ctext(message)
        super().__init__(self.message)


class MissingBrowsermobProxyPath(Exception):
    # Raise this when browsermob proxy path is missing
    def __init__(
        self,
        message="<y>'browsermob_proxy_path' is empty. Check your proxy configuration.",
    ):
        self.message = ctext(message)
        super().__init__(self.message)


class SeleniumWireAndUndetectedChromedriverIncompatible(Exception):
    # Raise this when undetected_chromedriver is not installed
    def __init__(
        self,
        message="<y>Selenium Wire and Undetected Chromedriver are incompatible. Please only set 'use_selenium_wire' OR 'use_undetected_chromedriver'.",
    ):
        self.message = ctext(message)
        super().__init__(self.message)


class SeleniumWireNotInstalled(Exception):
    # Raise this when undetected_chromedriver is not installed
    def __init__(
        self,
        message="<y>Module 'seleniumwire' not installed.",
    ):
        self.message = ctext(message)
        super().__init__(self.message)


class UndetectedChromedriverNotInstalled(Exception):
    # Raise this when undetected_chromedriver is not installed
    def __init__(
        self,
        message="<y>Module 'undetected_chromedriver' not installed.",
    ):
        self.message = ctext(message)
        super().__init__(self.message)



# Standard Libraries
from base64 import b64decode
import json
import time

# Third-party Libraries
from lytils import cprint
from lytils.file import LyFile
from lytils.regex import match
from selenium.common.exceptions import TimeoutException

# * OPTIONAL IMPORTS
try:
    from browsermobproxy import Server
except ModuleNotFoundError:
    Server = None

try:
    import psutil  # type: ignore
except ModuleNotFoundError:
    psutil = None

# Local Libraries
from .exceptions import BrowsermobProxyNotInstalled, MissingBrowsermobProxyPath


class BrowsermobProxyWrapper:
    def __init__(self, browsermob_proxy_path: str, port=9090, debug: bool = False):
        self._debug = debug
        if Server is None:
            raise BrowsermobProxyNotInstalled
        if not browsermob_proxy_path:
            raise MissingBrowsermobProxyPath

        options = {"port": port}

        if self._debug:
            cprint(f"<y>DEBUG: BMP browsermob_proxy_path = {browsermob_proxy_path}")
            cprint(f"<y>DEBUG: BMP port = {port}")
            options.update({"log_path": "browsermob-proxy.log"})

        self.kill_existing_bmp()  # Kill existing browsermob proxies before starting
        self._server = Server(browsermob_proxy_path, options=options)
        self._client = None

    def kill_existing_bmp(self):
        for proc in psutil.process_iter():
            # check whether the process name matches
            if proc.name() == "browsermob-proxy":
                proc.kill()

    def start_server(self):
        self._server.start()
        return self._server

    def start_client(self):
        # Disable certificate checks. Proxy will trust all of the servers.
        # ! Only use "trustAllServers" with crawling/testing, unsafe to use for browsing.
        # params = { "trustAllServers": "true" }
        params = {}
        self._client = self._server.create_proxy(params=params)
        return self._client

    def server(self):
        return self._server

    def client(self):
        return self._client

    def proxy_url(self):
        return self._client.proxy

    def start_har(self, name: str | None = None):
        # Possible HAR options
        # Source: https://medium.com/@jiurdqe/how-to-get-json-response-body-with-selenium-amd-browsermob-proxy-71f10335c66
        options = {
            "captureHeaders": True,
            "captureContent": True,
            "captureBinaryContent": True,
        }
        self._client.new_har(name, options=options)

    def har(self):
        return self._client.har

    def _get_response_from_content(self, content: dict, debug: bool = False):
        debug_path = "debug/BMP/_get_response_from_content.txt"

        try:
            # No decoding necessary
            return json.loads(content["text"])

        except json.decoder.JSONDecodeError:
            # response -> content -> text is not a json string
            encoding = content["encoding"]

            if encoding == "base64":
                charset = match(r"(?<=charset=).+", content["mimeType"])
                decoded_bytes = b64decode(content["text"])

                # decoded_str = decoded_bytes.decode(charset)
                decoded_str = decoded_bytes.decode(charset)

                if debug:
                    # Print request url and response
                    append_line(debug_path, decoded_str)

                return json.loads(decoded_str)

        except KeyError:
            # response -> content -> text does not exist
            raise KeyError

    def wait_for_response(
        self,
        partial_request_url: str,
        poll_interval: int = 1,
        timeout: int = 10,
        debug: bool = False,
    ):
        """
        Either returns a response or raises a timeout exception.
        """
        debug_file = LyFile("debug/BMP/wait_for_response.txt")
        if debug:
            debug_file.create()

        start_time = time.time()  # Record start time for timeout

        while time.time() - start_time < timeout:

            for entry in self._client.har["log"]["entries"]:

                if partial_request_url in json.dumps(entry["request"]["url"]):

                    if debug:
                        # Print request url and response
                        debug_file.append(entry["request"]["url"])
                        debug_file.append_json(entry["response"])

                    content = entry["response"]["content"]

                    try:
                        # No decoding necessary
                        return self._get_response_from_content(content, debug=True)

                    except KeyError:
                        # response -> content -> text does not exist
                        continue

            # ping requests against after interval has passed
            time.sleep(poll_interval)

        # If code reaches outside of loop, it is because of timeout.

        if debug:
            duration = time.time() - start_time
            debug_file.append(f"Timed out. Duration: {round(duration, 3)} seconds.")

        raise TimeoutException

    # Troubleshoot request urls
    def get_request_urls(self):
        request_urls = []
        for entry in self._client.har["log"]["entries"]:
            request_urls.append(entry["request"]["url"])
        return request_urls

    def get_responses(self, partial_request_url: str, debug: bool = False):
        debug_file = LyFile("debug/BMP/get_responses.txt")
        if debug:
            debug_file.create()

        responses = []
        for entry in self._client.har["log"]["entries"]:

            if partial_request_url in entry["request"]["url"]:

                if debug:
                    # Print request url and response
                    debug_file.append(entry["request"]["url"])
                    debug_file.append_json(entry["response"])

                content = entry["response"]["content"]

                try:
                    response = self._get_response_from_content(content, debug=False)
                    responses.append(response)

                except KeyError:
                    continue

        if debug and len(responses) == 0:
            debug_file.append(f"url '{partial_request_url}' not found.")

        return responses

    def close(self):
        self._client.close()
        self._server.stop()



from typing import Callable

# Third-party libraries
from gspread_formatting import CellFormat
from lytils import cprint
from lytils.regex import match
from lytils.google_sheets import Column as GSColumn
from lytils.google_sheets.format import DefaultFormat


# GSColumn is a class used to configure GoogleSheet columns
# This Column class extends GSColumn to support columns for crawling
class Column(GSColumn):
    def __init__(
        self,
        header: str,
        xpath: str = "",
        regex: str | list[str] = "",
        transform: Callable[[str], str] | None = None,
        format: CellFormat = DefaultFormat(),
        width: int = 100,
    ):
        # Initialize GoogleSheet Column instance
        super().__init__(header=header, format=format, width=width)

        # Xpath used to locate element on webpage
        self._xpath = xpath

        # Allow for single regex or list of regex
        self._regex = regex

        # Last line of defense for manipulating the crawled data
        #   is calling a function to do so
        # Locating by xpath and applying regex should be more
        #   than enough for most scenarios
        self._transform = transform

    def has_xpath(self):
        return self._xpath != ""

    def get_xpath(self):
        return self._xpath

    def has_regex(self):
        return self._regex != ""

    def apply_regex(self, text: str) -> str:
        new_text = text
        if isinstance(self._regex, str):
            try:
                new_text = match(self._regex, new_text)
            except:
                cprint(f"<y>RegEx Error ::: regex: {self._regex}, text: {text}")
        else:
            for r in self._regex:
                try:
                    new_text = match(r, new_text)
                except:
                    cprint(f"<y>RegEx Error ::: regex: {r}, text: {text}")
                    break
        return new_text

    def has_transform(self):
        return self._transform != None

    # Takes a string and calls the transform function, returning
    # the transformed string
    def transform(self, text: str) -> str:
        return self._transform(text)



# Base Python Libraries
import asyncio

# Third-Party Libraries
from playwright.async_api import async_playwright
from playwright.async_api import Browser, BrowserContext, Page
from playwright.async_api import Route, Request

# Local Libraries
from qrawl import add_params_to_url, get_cookie_header, get_random_user_agent


async def get_new_context(
    browser: Browser,
    user_agent: str = "",
    headers: dict | None = None,
    cookies: dict | None = None,
) -> BrowserContext:
    # Get random user agent if not defined
    if user_agent == "":
        user_agent = get_random_user_agent()

    context = await browser.new_context(user_agent=user_agent)

    h = {
        # Define default headers here
        "referer": "https://www.google.com/",
        "user-agent": user_agent,
    }

    # Add additional headers
    if headers is not None:
        h.update(headers)

    # Add cookies
    if cookies is not None:
        await context.add_cookies(cookies)
        h.update({"cookies": get_cookie_header(cookies)})

    await context.set_extra_http_headers(h)

    return context


async def get_new_page(context: BrowserContext) -> Page:
    page = await context.new_page()

    # Set webdriver to false automatically for added stealth
    await set_webdriver_to_false(page)
    return page


async def set_webdriver_to_false(page: Page):
    """
    Sets navigator.webdriver to false to avoid detection.
    """
    await page.evaluate(
        "Object.defineProperty(navigator, 'webdriver', { get: () => false })"
    )


async def visit_url_with_params(page: Page, url: str, params: dict):
    url = add_params_to_url(url, params)
    return await page.goto(url)


async def visit_google_cache(page: Page, url: str, params: dict = {}):
    """
    https://webcache.googleusercontent.com/search?q=cache:<url>
    If a webpage is too heavily protected and frequent updates are unnecessary,
    consider scraping google's cached version of the page instead.
    """
    CACHE_URL = "https://webcache.googleusercontent.com/search?q=cache"
    url = add_params_to_url(url, params)
    return await page.goto(f"{CACHE_URL}:{url}")


async def block_resources(
    page: Page,
    resources: list[str] = [
        "image",
        "stylesheet",
        "media",
        "font",
        "other",
    ],
):
    """
    To improve efficiency, exclude unnecessary elements from loading.
    See https://scrapingant.com/blog/block-requests-playwright
    """

    async def route_guard(route: Route, request: Request):
        # Block requests for excluded resource types
        if request.resource_type in resources:
            await route.abort()
        else:
            await route.continue_()

    await page.route("**/*", route_guard)


class QAsyncPlaywright:
    def __init__(self, max_concurrency: int = 5):
        self._p = None
        self._initialized = False
        self._semaphore = asyncio.Semaphore(max_concurrency)

    def get_semaphore(self):
        return self._semaphore

    async def initialize(self):
        if not self._initialized:
            self._p = await async_playwright().start()
            self._initialized = True

    async def launch_browser(self, *args, **kwargs):
        return await self._p.chromium.launch(*args, **kwargs)

    async def launch_persistent_context(self, path: str, *args, **kwargs):
        """
        For launching a persistent browser from a given path.
        The path must lead to an existing folder.
        Once launched for the first time, the folder will populate with a session.
        """
        return await self._p.chromium.launch_persistent_context(
            user_data_dir=path, *args, **kwargs
        )

    async def close(self):
        if self._initialized:
            await self._p.stop()
            self._initialized = False
            print("Playwright closed.")
